from django import forms
from django.contrib.auth import get_user_model
from django.contrib.auth.views import REDIRECT_FIELD_NAME, LoginView, SuccessURLAllowedHostsMixin
from django.views.generic.edit import FormView

from swap_user.helpers import normalize_username
from swap_user.otp.forms import CheckOTPForm, GetOTPForm
from swap_user.settings import swap_user_settings


UserModel = get_user_model()


class GetOTPView(FormView):
    """
    This view helps to send OTP via `OTPSender` subclasses
    """

    template_name = "admin/login-otp.html"
    form_class = GetOTPForm
    success_url = "/admin/check-otp/?next=/admin/"

    def form_valid(self, form: forms.Form):
        """
        If form is valid, we will use corresponding service and do the following:
            - Generate OTP and send it
            - Save username for autofill on the next screen
        """

        username_field = UserModel.USERNAME_FIELD
        raw_username = form.cleaned_data[username_field]
        username = normalize_username(raw_username)
        request = self.request

        service_class = swap_user_settings.GET_OTP_SERVICE_CLASS
        service = service_class()
        service.generate_otp_and_send(username=username)
        service.save_username_to_sesson(request=request, username=username)

        return super().form_valid(form)


class CheckOTPView(SuccessURLAllowedHostsMixin, FormView):
    """
    We are validating OTP that comes from user.
    """

    template_name = "admin/check-otp.html"
    form_class = CheckOTPForm
    redirect_field_name = REDIRECT_FIELD_NAME

    def get_initial(self):
        """
        Let's try to autofill username at the next screen after OTP was sent.
        Convenient for human - no need to provide email one more time.
        """

        session = self.request.session
        cached_username = session.get(UserModel.USERNAME_FIELD, "")
        username_field = UserModel.USERNAME_FIELD

        initial = {username_field: cached_username}

        return initial

    def form_valid(self, form: forms.Form):
        """
        If form is valid - we will do:
            - Authenticate
            - Login
        """

        username_field = UserModel.USERNAME_FIELD
        raw_username = form.cleaned_data[username_field]
        username = normalize_username(raw_username)
        otp_password = form.cleaned_data["otp"]

        service_class = swap_user_settings.CHECK_OTP_SERVICE_CLASS
        service = service_class()
        service.authenticate_and_login(
            request=self.request, username=username, password=otp_password,
        )

        return super().form_valid(form)

    def form_invalid(self, form):
        """
        If form is invalid - we will increase counter of invalid logins for current user.
        When counter will reach the limit - user will be banned for some amount of time.
        """

        username_field = UserModel.USERNAME_FIELD
        raw_username = form.cleaned_data[username_field]
        username = normalize_username(raw_username)

        service_class = swap_user_settings.CHECK_OTP_SERVICE_CLASS
        service = service_class()
        service.track_invalid_login_attempt(username=username)

        return super().form_invalid(form)

    get_success_url = LoginView.get_success_url
    get_redirect_url = LoginView.get_redirect_url



from django.contrib.admin import apps


class DjangoSwapUser(apps.AppConfig):
    """
    This is default app config
    """

    name = "swap_user"
    verbose_name = "Django Swap User"


class OTPSiteConfig(apps.AdminConfig):
    """
    This config required for customizing admin panel Login form.
    """

    # custom admin site with customized Login form.
    default_site = "swap_user.otp.sites.OTPUserSite"



from abc import ABC, abstractmethod


class AbstractOTPSender(ABC):
    """
    Abstract sender class interface.
    Subclass this and implement following methods:
        - `send`
        - `_render_message`
    """

    @abstractmethod
    def send(self, receiver: str, otp: str, **kwargs):
        """
        Put whole sending logic into this method.
        """

        pass

    @abstractmethod
    def _render_message(self, otp: str, **kwargs):
        """
        Here we can add some formatting / styling for message that you
        are going to send.
        """

        pass



from swap_user.base_forms import BaseUserOptionalFieldsForm, BaseUserRequiredFieldsForm
from swap_user.to_email.models import EmailUser


class EmailUserOptionalFieldsForm(BaseUserOptionalFieldsForm):
    """
    Form for - EmailUser model.
    With optional `password_1` and `password_2` fields.
    """

    class Meta(BaseUserOptionalFieldsForm.Meta):
        model = EmailUser


class EmailUserRequiredFieldsForm(BaseUserRequiredFieldsForm):
    """
    Form for - EmailUser model.
    With required `password_1` and `password_2` fields.
    """

    class Meta(BaseUserRequiredFieldsForm.Meta):
        model = EmailUser



import logging

from django.conf import settings
from django.core.mail import send_mail
from django.utils.translation import gettext_lazy as _

from swap_user.base_senders import AbstractOTPSender


logger = logging.getLogger(__name__)


class StdOutOTPSender(AbstractOTPSender):
    """
    Simple sender that just writes all messages to STDOUT.
    """

    def send(self, receiver: str, otp: str, **kwargs):
        """
        Writes OTP message to STDOUT.
        """

        message = self._render_message(otp)
        logger.info(message)

    def _render_message(self, otp: str, **kwargs):
        return f"OTP: {otp}"


class EmailOTPSender(AbstractOTPSender):
    """
    Sender that implements sending OTP via default `send_mail`
    function. To make this work you need to set following settings:
        - `EMAIL_HOST`
        - `EMAIL_HOST_PASSWORD`
        - `EMAIL_HOST_USER`
        - `EMAIL_PORT`
        - `EMAIL_USE_TLS`
        - `EMAIL_USE_SSL`
    """

    def send(self, receiver: str, otp: str, **kwargs):
        """
        Sends OTP via email.
        """

        subject = _("OTP")
        message = self._render_message(otp)
        from_email = self._get_from_email()
        recipient_list = [receiver]

        send_mail(
            subject=subject, message=message, from_email=from_email, recipient_list=recipient_list,
        )

    def _render_message(self, otp: str, **kwargs):
        return f"OTP: {otp}"

    def _get_from_email(self):
        """
        Override this, if your settings are differs from this.
        """

        return settings.EMAIL_HOST_USER



from typing import Optional

from django.contrib.auth import authenticate, get_user_model, login
from django.http import HttpRequest

from swap_user.helpers import (
    generate_otp,
    get_banned_user_cache_key,
    get_invalid_login_cache_key,
    get_otp_cache_key,
    increase_counter_of_invalid_login,
    set_key_to_cache,
)
from swap_user.settings import swap_user_settings


UserModel = get_user_model()
USER_IS_BANNED = True


class GetOTPService:
    """
    Service for `GetOTPView`, that handles whole logic of sending OTP
    code such as:
        - Decides we can send OTP to this User or not
        - Generates and caches OTP
        - Sends OTP to User
    """

    def generate_otp_and_send(self, username: str):
        """
        Main handler of service, which holds all logical steps.
        """

        user = self._get_user(username)
        if not user:
            return None

        is_enough_permissions = self._has_enough_permissions(user)
        if not is_enough_permissions:
            return None

        user_id = user.id
        sender_class = swap_user_settings.OTP_SENDER_CLASS

        otp = generate_otp()
        cache_key = get_otp_cache_key(user_id)
        set_key_to_cache(cache_key=cache_key, value=otp)

        sender = sender_class()
        sender.send(username, otp)

    def save_username_to_sesson(self, request: HttpRequest, username: str):
        """
        Save username to session for future usage at the next
        screen (view `CheckOTPView`). Just convenient for future steps.
        """

        request.session[UserModel.USERNAME_FIELD] = username

    def _get_user(self, username: str) -> Optional[UserModel]:
        """
        Method, that handles user presence in our DB or not.
        """

        username_field = UserModel.USERNAME_FIELD
        query_data = {username_field: username}

        try:
            user = UserModel.objects.get(**query_data)
        except UserModel.DoesNotExist:
            return None

        return user

    def _has_enough_permissions(self, user: UserModel) -> bool:
        """
        Override this, if you need to customize conditions when email
        will be sent to user.
        """

        is_staff = user.is_staff
        is_active = user.is_active

        return is_staff and is_active


class CheckOTPService:
    """
    Service for `CheckOTPView` which authenticates (or not) and then
    we are proceeding through default Django's login process - i.e. writing
    to session and cookies.
    """

    def authenticate_and_login(self, request: HttpRequest, username: str, password: str):
        """
        Default authentication and login process to enter Django's admin.
        """

        user = authenticate(request, username=username, password=password,)
        login(request, user)

    def track_invalid_login_attempt(
        self,
        username: str,
        max_invalid_attempts: int = swap_user_settings.MAX_INVALID_LOGIN_ATTEMPTS,
        ban_timeout: int = swap_user_settings.BANNED_USER_TIMEOUT,
    ):
        """
        Here we are going to track all invalid login attempts.
        When invalid attempts will reach a limit - user will be banned for some period.
        """

        invalid_login_cache_key = get_invalid_login_cache_key(username)
        current_counter = increase_counter_of_invalid_login(invalid_login_cache_key)

        if current_counter < max_invalid_attempts:
            return None

        banned_user_cache_key = get_banned_user_cache_key(username)
        set_key_to_cache(
            cache_key=banned_user_cache_key, value=USER_IS_BANNED, expire=ban_timeout,
        )



from typing import Optional

from django.contrib import admin
from django.http import HttpRequest
from django.urls import path
from django.utils.decorators import method_decorator
from django.views.decorators.cache import never_cache
from django.views.decorators.csrf import csrf_protect
from django.views.decorators.debug import sensitive_post_parameters

from swap_user.otp.views import CheckOTPView, GetOTPView


class OTPUserSite(admin.AdminSite):
    def get_urls(self):
        """
        Here we are adding a new route /check-otp/ to the parent routes.
        """

        default_urls = super().get_urls()

        custom_urls = [
            path("check-otp/", self.check_otp, name="check-otp"),
        ]

        # Order is matter
        urls = custom_urls + default_urls

        return urls

    @method_decorator(sensitive_post_parameters())
    @method_decorator(csrf_protect)
    @never_cache
    def login(self, request: HttpRequest, extra_context: Optional[dict] = None):
        """
        At this view handler we are registering custom `GetOTPView`
        which sends an OTP to user via provided sender.
        """

        request.current_app = self.name
        context = self._get_context(request, extra_context)

        return GetOTPView.as_view(**context)(request)

    @method_decorator(sensitive_post_parameters())
    @method_decorator(csrf_protect)
    @never_cache
    def check_otp(self, request: HttpRequest, extra_context: Optional[dict] = None):
        """
        This view checks received OTP with OTP cached at backend side.
        """

        request.current_app = self.name
        context = self._get_context(request, extra_context)

        return CheckOTPView.as_view(**context)(request)

    def _get_context(self, request, extra_context: dict) -> dict:
        """
        Let's create a context for view
        Ref - django.contrib.admin.sites#login
        """

        context = {
            **self.each_context(request),
            **(extra_context or {}),
        }
        defaults = {
            "extra_context": context,
        }

        return defaults
